/*****************************************************************************\
 *  slurm_protocol_util.h - communication infrastructure functions
 *****************************************************************************
 *  Copyright (C) 2002 The Regents of the University of California.
 *  Produced at Lawrence Livermore National Laboratory (cf, DISCLAIMER).
 *  Written by Kevin Tew <tew1@llnl.gov> et. al.
 *  CODE-OCEC-09-009. All rights reserved.
 *
 *  This file is part of Slurm, a resource management program.
 *  For details, see <https://slurm.schedmd.com/>.
 *  Please also read the included file: DISCLAIMER.
 *
 *  Slurm is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  In addition, as a special exception, the copyright holders give permission
 *  to link the code of portions of this program with the OpenSSL library under
 *  certain conditions as described in each individual source file, and
 *  distribute linked combinations including the two. You must obey the GNU
 *  General Public License in all respects for all of the code used other than
 *  OpenSSL. If you modify file(s) with this exception, you may extend this
 *  exception to your version of the file(s), but you are not obligated to do
 *  so. If you do not wish to do so, delete this exception statement from your
 *  version.  If you delete this exception statement from all source files in
 *  the program, then also delete it here.
 *
 *  Slurm is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 *  details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with Slurm; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA.
\*****************************************************************************/

#ifndef _SLURM_PROTOCOL_UTIL_H
#define _SLURM_PROTOCOL_UTIL_H

#include <inttypes.h>

#include "src/common/pack.h"
#include "src/common/slurm_protocol_defs.h"

/*
 * init_header - simple function to create a header, always insuring that
 * an accurate version string is inserted
 * OUT header - the message header to be send
 * IN msg_type - type of message to be send
 * IN flags - message flags to be send
 */
extern void
init_header(header_t * header, slurm_msg_t *msg, uint16_t flags);


/*
 * update_header - update a message header with the message len
 * OUT header - the message header to update
 * IN msg_length - length of message to be send
 */
extern void
update_header(header_t * header, uint32_t msg_length);

/* Get the port number from a slurm_addr_t */
extern uint16_t slurm_get_port(slurm_addr_t *addr);

/* Set the port number in a slurm_addr_t */
extern void slurm_set_port(slurm_addr_t *addr, uint16_t port);

/* Check if slurm_addr_t has an unspecified address type */
extern bool slurm_addr_is_unspec(slurm_addr_t *addr);

typedef enum {
	RPC_FINGERPRINT_INVALID = 0,
	RPC_FINGERPRINT_FOUND,
	RPC_FINGERPRINT_NOT_FOUND,
	RPC_FINGERPRINT_NEED_MORE_BYTES,
	RPC_FINGERPRINT_INVALID_MAX,
} rpc_fingerprint_t;

/*
 * Fingerprint buffer for RPC
 * IN buffer - buffer to check
 * RET fingerprinting status
 */
extern rpc_fingerprint_t rpc_fingerprint(const buf_t *buffer);

#endif /* !_SLURM_PROTOCOL_UTIL_H */
